
import glfw
import OpenGL.GL as gl

import imgui
from imgui.integrations.glfw import GlfwRenderer
import time


def main(name, widget, width, height, maximized=False, bg_color=(0.9, 0.9, 0.9), pass_window_to_widget=False):
    """ Create a GLFW window, spin up the main loop, and display a given widget inside.

    If `pass_window_to_widget` is `True`, the `widget` parameter must be a function which takes a GLFW window handle
    and returns a widget. Else, `widget` is just a widget. This is useful if the widget should, for example, scale
    with the GLFW window.
    """
    imgui.create_context()

    # Set config flags
    imgui.get_io().config_flags |= imgui.CONFIG_DOCKING_ENABLE | imgui.CONFIG_VIEWPORTS_ENABLE

    window = create_window(name, width, height, maximized)
    impl = GlfwRenderer(window)
    if pass_window_to_widget:
        widget = widget(window)

    ## Using this feels significantly choppier than sleeping manually. TODO: investigate & fix
    # glfw.swap_interval(-1)

    while not glfw.window_should_close(window):
        t0 = time.perf_counter()
        glfw.poll_events()
        impl.process_inputs()

        imgui.new_frame()

        try:
            next(widget)
        except StopIteration:
            imgui.render()
            break
        except:
            # Cleanup on exception for iPython
            imgui.render()
            glfw.terminate()
            raise

        gl.glClearColor(*bg_color, 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        imgui.render()
        impl.render(imgui.get_draw_data())
        glfw.swap_buffers(window)

        t1 = time.perf_counter()
        if t1 - t0 < 1/60:
            time.sleep(1/60 - (t1 - t0))

    impl.shutdown()
    glfw.terminate()


def create_window(window_name, width, height, maximized):
    """ Create a GLFW window. """
    if not glfw.init():
        print("Could not initialize OpenGL context")
        exit(1)

    # OS X supports only forward-compatible core profiles from 3.2
    glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
    glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
    glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)

    glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)
    if maximized:
        glfw.window_hint(glfw.MAXIMIZED, gl.GL_TRUE)

    # Create a windowed mode window and its OpenGL context
    window = glfw.create_window(
        int(width), int(height), window_name, None, None
    )
    glfw.make_context_current(window)

    if not window:
        glfw.terminate()
        print("Could not initialize Window")
        exit(1)

    return window



""" Experimental widget for image sequence or video seeking. """


import concur as c


class TimeControls(object):
    """ State for the `concur.extra_widgets.time_controls.time_controls` widget. """
    def __init__(self, start, end, hotkey=None):
        """ Initialize. """
        self.frame_id = 0.0
        self.start = 0
        self.end = end - 1
        self.playing = False
        self.stride = 10
        self.hotkey = hotkey

def time_controls(state):
    """ Experimental widget for image sequence or video seeking.

    Takes `concur.extra_widgets.time_controls.TimeControls` as an argument. Returns a pair `(state, frame_id)`, where `state`
    is the modified `TimeControls` struct, and `frame_id` is an integer signalling a new position in the video frame,
    or `None` if user didn't seek.
    """
    def clock():
        return "tick", None
        yield

    tag, value = yield from c.window("Time Controls",
        [ c.slider_int("Frame ID", state.frame_id, state.start, state.end, tag="seek")
        , c.slider_float("Local Seek", 0, -1.5, 1.5, tag="drag")
        , c.slider_int("Stride", state.stride, 1, 20, tag="stride")
        , c.orr_same_line(
            [ c.button("Next")
            , c.button("Prev")
            , c.button("Pause" if state.playing else "Play", tag="PlayPause")
            ])
        ]
         + ([] if not state.hotkey else [c.key_pressed("space", state.hotkey, True)])
         + ([] if not state.playing else [clock()]))

    last_frame_id = state.frame_id
    if tag == "seek":
        state.frame_id = value
    elif tag == "drag":
        state.frame_id += 10 ** value - 10 ** (-value)
    elif tag == "stride":
        state.stride = value
    elif tag == "Next":
        state.frame_id += 1
    elif tag == "Prev":
        state.frame_id -= 1
    elif tag == "PlayPause":
        state.playing = not state.playing
    elif tag == "space":
        state.playing = not state.playing
    elif tag == "tick":
        state.frame_id += state.stride
    else:
        raise ValueError(f"Unknown event tag: {tag}")


    state.frame_id = min(state.end, max(state.start, state.frame_id))
    if int(last_frame_id) != int(state.frame_id):
        event = int(state.frame_id)
    else:
        event = None
    return state, event



""" Functions providing core functionality, widget creation, manipulation, and composition.

All of the functions in this module are re-exported in the root module for convenience.
"""


import queue
from typing import Generator, Any, Iterable, List, Callable, Tuple
from asyncio import Future
from imgui import push_id, pop_id


Widget = Any # It isn't possible to type Widget correctly using mypy. Prove me wrong.


def orr(widgets: Iterable[Widget]) -> Widget:
    """ Chain elements in space, returning the first event fired. """
    stop = False
    value = None
    while True:
        for i, elem in enumerate(widgets):
            try:
                push_id(str(i))
                next(elem)
            except StopIteration as e:
                if not stop:
                    stop = True
                    value = e.value
            finally:
                pop_id()
        if stop:
            return value
        else:
            yield


def multi_orr(widgets: Iterable[Widget]) -> Widget:
    """ Chain elements in space, returning all the events fired as a list """
    events: List = []
    while events == []:
        for i, elem in enumerate(widgets):
            try:
                push_id(str(i))
                next(elem)
            except StopIteration as e:
                events.append(e.value)
            finally:
                pop_id()
        if events == []:
            yield
    return events


def forever(elem_gen: Callable[..., Widget], *args, **kwargs) -> Widget:
    """ Repeat an element forever.

    Function generating the element must be passed as the first argument;
    remaining arguments are passed to said function.
    """
    elem = elem_gen(*args, **kwargs)
    while True:
        try:
            next(elem)
        except StopIteration:
            elem = elem_gen(*args, **kwargs)
        yield



def lift(f: Callable[..., Any], *args, **argv) -> Widget:
    """ Lift a function into a never-ending widget.

    Useful for wrapping ImGui calls and passive widgets.
    It is used to wrap some ImGui calls in `concur.widgets`.
    """
    while True:
        f(*args, **argv)
        yield


def interactive_elem(elem, name, *args, **kwargs):
    """ Function useful for wrapping a wide range of ImGui widgets.

    Elements which take `name` as the first argument and return
    a pair `(changed, value)` can be wrapped using this function.
    It is used to wrap some imgui widgets in `concur.widgets`.
    """
    if 'tag' in kwargs:
        tag = kwargs['tag']
        del kwargs['tag']
    else:
        tag = name
    while True:
        changed, value = elem(name, *args, **kwargs)
        if changed:
            return tag, value
        else:
            yield


def nothing():
    """ Widget that does nothing. """
    while True: yield


def event(ev):
    """ Widget that immediately returns `ev`. """
    return ev
    yield


def optional(exists: bool, widget_gen: Callable[..., Widget], *args, **kwargs):
    """ Optionally display a widget. """
    return widget_gen(*args, **kwargs) if exists else nothing()


def tag(tag_name: Any, elem: Widget) -> Widget:
    """ Transform any returned value `v` of `elem` into a tuple `tag_name, v`. """
    return tag_name, (yield from elem)


def map(f: Any, elem: Widget) -> Widget:
    """ Transform any returned value `v` of `elem` into `f(v)`. """
    v = yield from elem
    return f(v)


def stateful(elem: Callable[[Any], Widget], initial_state: Any) -> Widget:
    state = initial_state
    while True:
        state = yield from elem(state)
        yield


class Block(object):
    """ Create a widget that returns on future result. Useful for async computations.

    This widget is constructed manually using a class, because the future must be
    canceled in the destructor. Destructor isn't available in generator functions.
    """
    def __init__(self, future):
        self.future = future

    def __iter__(self):
        return self

    def __next__(self):
        if self.future.done():
            raise StopIteration(self.future.result())

    def __del__(self):
        self.future.cancel()


def listen(que):
    """ Listen to messages on a given queue. """
    while True:
        try:
            return que.get_nowait()
        except queue.Empty:
            yield


class RemoteAction(object):
    def __init__(self, future, que):
        self.que = que
        self.future = future
        self.sent = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.future is not None and not self.sent and self.future.done():
            self.que.put(self.future.result())
            self.sent = True

    def __del__(self):
        if self.future is not None:
            self.future.cancel()


def remote_widget(future):
    """ Separate the effect of the widget from its result """
    q = queue.Queue()

    def value():
        while True:
            try:
                v = q.get_nowait()
                return v
            except queue.Empty:
                yield

    return RemoteAction(future, q), value


def fork_action(future, rest_gen):
    """ A common pattern - running a long running action (`future`) and keeping the GUI (`rest`) responsive.

    Because the action can't be restarted on every gui event, we must *fork* it off in the beginning.
    """
    action, value_gen = remote_widget(future)
    return orr([action, rest_gen(value_gen)])
